// Copyright (C) 1999 by Makoto Nagata, MicroNova <mnagata@micronova.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit http://www.micronova.com for updates

package com.micronova.mnfc;

import netscape.application.*;
import netscape.util.*;

/**

   IToolTipManager manages tooltips.
   Typically an IToolTipManager is attached to an IApplication.

   * @author Makoto Nagata, MicroNova
   * @version 1.01

*/

public class IToolTipManager implements Target
{
    /** Delay timer command. */

    public static final String TIMER = "com.micronova.mnfc.IToolTipManager.TIMER";

    /** Idle state (the mouse is not in any tooltip view). */

    public static final int STATE_IDLE = 0;

    /** The Mouse has entered a tooltip view and waiting for initialDelay timer. */

    public static final int STATE_INITWAIT = 1;

    /** The tooltip is shown and waiting for dismissDelay timer. */

    public static final int STATE_TIPSHOWN = 2;

    /** The tooltip is dismissed but the mouse is still inside the tooltip view. */

    public static final int STATE_TIPDISMISSED = 3;

    /** The mouse has exited a tooltip view and waiting for reshowDelay timer. */

    public static final int STATE_RESHOWWAIT = 4;

    /** Holds registered IToolTips. */

    public Hashtable table = new Hashtable();

    /** The delay before which a tooltip will appear after the mouse enters a tooltip view. */

    public int initialDelay = 500;

    /** The delay before the current tooltip is dismissed if the mouse pointer does not move. */

    public int dismissDelay = 4000;

    /** The delay within which a tooltip will be reshown at once if the mouse pointer exits and enters another tooltip view. */

    public int reshowDelay = 500;

    /** Set to true if this IToolTipManager is enabled. */

    public boolean enabled = true;

    /** Delay timer. */

    public Timer timer;

    /** Set to true if the View under the mouse pointer is an instance of IToolTipSource (for dynamic tooltip). */

    public boolean isToolTipSource = false;

    /** Current state. */

    public int state = STATE_IDLE;

    /** InternalWindow that displays tooltip. */

    public InternalWindow toolTipWindow;

    /** InternalWindow layer for the tooltip window. */

    public int toolTipWindowLayer = InternalWindow.POPUP_LAYER;

    /** Default TextView that displays tooltip text. */

    public TextView toolTipView;

    /** Default TextView background color. */

    public Color toolTipBackgroundColor = new Color(255, 255, 192);

    /** Default TextView foreground color. */

    public Color toolTipForegroundColor = Color.black;

    /** Default TextView font. */

    public Font toolTipFont = Font.defaultFont();

    /** Default tooltip border. */

    public Border toolTipBorder = LineBorder.blackLine();

    /** X-offset from the mouse pointer for the tooltip window. */

    public int xOffset = 10;

    /** Y-offset from the mouse pointer for the tooltip window. */

    public int yOffset = 10;

    /** Default tooltip window width. */

    public int toolTipWindowWidth = 200;

    /** Maximal number of text rows.  If the tooltip text exceeds this number of rows with the given toolTipWindowWidth, the window is expanded horizontally. */

    public int maxRows = 6;

    /** Current MouseEvent. */

    public MouseEvent mouseEvent;

    /** Current View under mouse pointer. */

    public View view;

    /** IToolTip to be displayed. */

    public IToolTip toolTip;

    /** 

        Registers an IToolTip.

    */

    public void registerToolTip(IToolTip toolTip)
    {
        table.put(toolTip.getView(), toolTip);
    }

    /**

       Unregisters an IToolTip.

    */
    
    public void unregisterToolTip(IToolTip toolTip)
    {
        table.remove(toolTip.getView());
    }

    /**

       Registers a tooltip 'toolTipText' for a 'view'.

    */

    public void registerToolTip(View view, String toolTipText)
    {
        registerToolTip(new IToolTip(view, toolTipText));
    }

    /**

       Unregisters tooltip for a view.

    */

    public void unregisterToolTip(View view)
    {
        table.remove(view);
    }

    /**

       Unregisters tooltips for all descendents of the given view (including the view itself).

    */

    public void unregisterToolTipForDescendents(View view)
    {
        unregisterToolTip(view);

        Vector subviews = view.subviews();

        for (int i = 0; i < subviews.size(); i ++)
        {
            View subview = (View)subviews.elementAt(i);

            unregisterToolTipForDescendents(subview);
        }
    }

    /**

       Returns the IToolTip registered for the view.  Returns null if there is no tooltip registered.

    */

    public IToolTip getToolTip(View view)
    {
        return (IToolTip)table.get(view);
    }

    /**

       Creates a View for tooltip.  The default implementation
       creates a TextView.  This is called once on initialization.

    */

    public View createToolTipView()
    {
        return new TextView();
    }

    /**

       Creates an InternalWindow for tooltip using the given toolTipView.  This is called once on initialization.  The tooltip window contains a ContainerView that holds toolTipView (created by createToolTipView()).

    */

    public InternalWindow createToolTipWindow(View toolTipView)
    {
        InternalWindow toolTipWindow = new InternalWindow();

        toolTipWindow.setType(InternalWindow.BLANK_TYPE);

        ContainerView containerView = new ContainerView();

        toolTipWindow.addSubview(containerView);

        containerView.addSubview(toolTipView);

        toolTipWindow.setLayer(toolTipWindowLayer);
        
        return toolTipWindow;
    }

    /**

       Constructs an IToolTipManager.

    */
    
    public IToolTipManager()
    {
        toolTipView = (TextView)createToolTipView();

        toolTipWindow = createToolTipWindow(toolTipView);

        timer = new Timer(this, TIMER, 0);

        timer.setRepeats(false);
    }

    /**

       Returns true if this tooltip manager is enabled.

    */

    public boolean isEnabled()
    {
        return enabled;
    }

    /**

       Enables/disables this tooltip manager.

    */

    public void setEnabled(boolean flag)
    {
        this.enabled = flag;
    }

    /**

       Sets the initial delay in microseconds.  This is the
       delay before which a tooltip will appear after the mouse
       enters a tooltip view.

    */

    public void setInitialDelay(int microSeconds)
    {
        initialDelay = microSeconds;
    }

    /**

       Returns current initial delay in microseconds.

    */

    public int getInitialDelay()
    {
        return initialDelay;
    }

    /**

       Sets the dismiss delay in microseconds.  This is the time
       a tooltip stays visible if the mouse does not move.

    */

    public void setDismissDelay(int microSeconds)
    {
        dismissDelay = microSeconds;
    }

    /**

       Returns the current dismiss delay in microseconds.

    */

    public int getDismissDelay()
    {
        return dismissDelay;
    }

    /**

       Sets the reshow delay in microseconds.  This is the time
       within which a tooltip will be reshown immediately
       when the mouse re-enters a tooltip view.

    */

    public void setReshowDelay(int microSeconds)
    {
        reshowDelay = microSeconds;
    }

    /**

       Returns the reshow delay in microseconds.

    */

    public int getReshowDelay()
    {
        return reshowDelay;
    }

    /**

       Prepares tooltip to be shown.

    */

    public void prepareToolTip(IToolTip toolTip)
    {
        String toolTipText = toolTip.getTipText();

        int x = 0;
        int y = 0;

        if (mouseEvent != null)
        {
            x = mouseEvent.x;
            y = mouseEvent.y;
        }

        toolTipWindow.moveTo(x + xOffset, y + yOffset);

        ((ContainerView)toolTipView.superview()).setBorder(toolTipBorder);

        toolTipView.setTextColor(toolTipForegroundColor);
        toolTipView.setBackgroundColor(toolTipBackgroundColor);
        toolTipView.setFont(toolTipFont);

        toolTipView.setString(toolTip.getTipText());

        // adjust window size

        int guessSize = toolTipWindowWidth;
        
        int maxSize = Application.application().mainRootView().width();
        
        if (guessSize > maxSize)
        {
            guessSize = maxSize;
        }

        while (guessSize <= maxSize)
        {
            toolTipView.sizeTo(guessSize, guessSize);
            
            toolTipView.sizeToMinSize();

            if (toolTipView.lineCount() <= maxRows)
            {
                break;
            }
            else
            {
                guessSize += 50;
            }
        }

        // minimize the window size, if possible

        int currentLineCount = toolTipView.lineCount();
        int minimalWidth = toolTipView.width();
        int absoluteMinimalWidth = toolTipFont.fontMetrics().charWidth('W');

        for (int width = toolTipView.width() - 1; width > absoluteMinimalWidth; width --)
        {
            toolTipView.sizeTo(width, toolTipView.height());
            toolTipView.sizeToMinSize();
                
            if (currentLineCount == toolTipView.lineCount())
            {
                minimalWidth = toolTipView.width();
            }
            else
            {
                break;
            }
        }
            
        toolTipView.sizeTo(minimalWidth, toolTipView.height());

        toolTipView.superview().sizeTo(toolTipView.width(), toolTipView.height());

        Size size = toolTipWindow.windowSizeForContentSize(toolTipView.width(), toolTipView.height());

        toolTipWindow.sizeTo(size.width, size.height);

        // try to make the window fully visible

        {
            int absoluteHeight = view.rootView().height();
            int absoluteWidth = view.rootView().width();

            Rect r = toolTipWindow.bounds();
            
            if (r.y > absoluteHeight - r.height)
            {
                r.y = absoluteHeight - r.height;
            }
            
            if (r.x > absoluteWidth - r.width)
            {
                r.x = absoluteWidth - r.width;
            }
            
            toolTipWindow.moveTo(r.x, r.y);
        }
    }

    /** 

        Shows the tooltip window.

    */

    public void showToolTip(IToolTip toolTip)
    {
        prepareToolTip(toolTip);
        toolTipWindow.show();
    }

    /**

       Hides the tooltip window.

    */

    public void hideToolTip()
    {
        toolTipWindow.hide();
    }

    /**

       Returns dynamic tooltip text for an IToolTipSource.

    */

    public String getToolTipText(IToolTip toolTip, MouseEvent event)
    {
        if (toolTip != null)
        {
            View view = toolTip.getView();

            MouseEvent mouseEventForSource = (view.rootView()).convertEventToView(view, event);

            return ((IToolTipSource)view).getToolTipText(mouseEventForSource);
        }
        else
        {
            return null;
        }
    }

    /**

       Starts the delay timer.

    */

    public void startTimer(int delay)
    {
        timer.setInitialDelay(delay);

        timer.start();
    }

    /**

       Stops the delay timer if running.

    */

    public void stopTimer()
    {
        if (timer.isRunning())
        {
            timer.stop();
        }
    }

    /**

       Called when the mouse enters a tooltip area of a view.

    */

    public void toolTipEntered(View view, MouseEvent event)
    {
        switch (state)
        {
            case STATE_IDLE:
            {
                timer.stop();

                this.view = view;
                this.mouseEvent = event;

                state = STATE_INITWAIT;

                startTimer(initialDelay);
            }
            break;

            case STATE_RESHOWWAIT:
            {
                stopTimer();

                this.view = view;
                this.mouseEvent = event;

                showToolTip(toolTip);

                state = STATE_TIPSHOWN;

                startTimer(dismissDelay);
            }
            break;
        }
    }

    /**

       Called when the mouse exits the tooltip area of the view.

    */

    public void toolTipExited(View view, MouseEvent event)
    {
        stopTimer();
        
        hideToolTip();

        if (isToolTipSource && (toolTip != null))
        {
            toolTip.setTipText(null);
        }

        if (event == null)
        {
            state = STATE_IDLE;
        }
        else
        {
            switch (state)
            {
                case STATE_INITWAIT:
                {
                    state = STATE_IDLE;
                }
                break;

                case STATE_TIPSHOWN:
                case STATE_TIPDISMISSED:
                {
                    if (state == STATE_TIPSHOWN)
                    {
                        state = STATE_RESHOWWAIT;
                    }
                    else
                    {
                        state = STATE_IDLE;
                    }

                    startTimer(reshowDelay);
                }
                break;
            }
        }
    }

    /**

       Called when the mouse enters a tooltip view.

    */

    public void mouseEntered(View view, MouseEvent event)
    {
        toolTip = getToolTip(view);

        if (toolTip != null)
        {
            isToolTipSource = (view instanceof IToolTipSource);

            if (isToolTipSource)
            {
                toolTip.setTipText(null);

                String newText = getToolTipText(toolTip, event);

                if (newText != null)
                {
                    toolTip.setTipText(newText);
                    toolTipEntered(view, event);
                }
            }
            else
            {
                toolTipEntered(view, event);
            }
        }
    }

    /**

       Called when the mouse exists a tooltip view.

    */

    public void mouseExited(View view, MouseEvent event)
    {
        toolTipExited(view, event);

        isToolTipSource = false;
    }

    /**

       Called when the mouse is moved within a tooltip view.

    */

    public void mouseMoved(View view, MouseEvent event)
    {
        this.mouseEvent = event;

        if (state == STATE_TIPSHOWN)
        {
            stopTimer();

            startTimer(dismissDelay);
        }

        if (isToolTipSource)
        {
            String currentTip = (toolTip != null) ? toolTip.getTipText() : null;
            String nextTip = getToolTipText(toolTip, event);

            if (currentTip == null)
            {
                if (nextTip != null)
                {
                    toolTip.setTipText(nextTip);
                    toolTipEntered(view, event);
                }
            }
            else
            {
                if (nextTip == null)
                {
                    toolTipExited(view, event);
                }
                else if (!nextTip.equals(currentTip))
                {
                    toolTipExited(view, event);
                    
                    toolTip.setTipText(nextTip);

                    toolTipEntered(view, event);
                }
            }
        }
    }

    /**

       Called when the delay timer expires.

    */

    public void timerUp()
    {
        switch (state)
        {
            case STATE_INITWAIT:
            {
                showToolTip(toolTip);

                state = STATE_TIPSHOWN;

                startTimer(dismissDelay);
            }
            break;
            
            case STATE_TIPSHOWN:
            {
                hideToolTip();
                state = STATE_TIPDISMISSED;
            }
            break;

            case STATE_RESHOWWAIT:
            {
                state = STATE_IDLE;
            }
            break;
        }
    }

    /**

       Overridden to handle timer events.
       
    */

    public void performCommand(String command, Object object)
    {
        if (command.equals(TIMER))
        {
            timerUp();
        }
    }
}

